package dev.fitko.fitconnect.client.bootstrap;

import com.nimbusds.jose.jwk.JWK;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.SenderConfig;
import dev.fitko.fitconnect.api.config.SubscriberConfig;
import dev.fitko.fitconnect.api.config.build.BuildInfo;
import dev.fitko.fitconnect.api.config.build.CertificateConfig;
import dev.fitko.fitconnect.api.config.defaults.SchemaConfig;
import dev.fitko.fitconnect.api.domain.schema.SchemaResources;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectInitialisationException;
import dev.fitko.fitconnect.api.services.auth.OAuthService;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.crypto.MessageDigestService;
import dev.fitko.fitconnect.api.services.events.CaseService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.events.SecurityEventService;
import dev.fitko.fitconnect.api.services.keys.KeyService;
import dev.fitko.fitconnect.api.services.reply.ReplyService;
import dev.fitko.fitconnect.api.services.routing.RoutingService;
import dev.fitko.fitconnect.api.services.schema.SchemaProvider;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.client.RouterClient;
import dev.fitko.fitconnect.client.SenderClient;
import dev.fitko.fitconnect.client.SubscriberClient;
import dev.fitko.fitconnect.client.sender.ReplyReceiver;
import dev.fitko.fitconnect.client.sender.SubmissionSender;
import dev.fitko.fitconnect.client.subscriber.ReplySender;
import dev.fitko.fitconnect.client.subscriber.SubmissionReceiver;
import dev.fitko.fitconnect.client.util.ValidDataGuard;
import dev.fitko.fitconnect.core.FitConnectDefaultService;
import dev.fitko.fitconnect.core.auth.DefaultOAuthApiService;
import dev.fitko.fitconnect.core.cases.CaseApiService;
import dev.fitko.fitconnect.core.cases.EventLogVerifier;
import dev.fitko.fitconnect.core.cases.SecurityEventTokenService;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.http.HttpClient;
import dev.fitko.fitconnect.core.http.interceptors.ApiRequestInterceptor;
import dev.fitko.fitconnect.core.http.interceptors.UserAgentInterceptor;
import dev.fitko.fitconnect.core.keys.PublicKeyApiService;
import dev.fitko.fitconnect.core.reply.ReplyApiService;
import dev.fitko.fitconnect.core.routing.RouteVerifier;
import dev.fitko.fitconnect.core.routing.RoutingApiService;
import dev.fitko.fitconnect.core.schema.SchemaResourceProvider;
import dev.fitko.fitconnect.core.submission.SubmissionApiService;
import dev.fitko.fitconnect.core.util.CertificateLoader;
import dev.fitko.fitconnect.core.validation.DefaultValidationService;
import okhttp3.Interceptor;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.yaml.snakeyaml.Yaml;

import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.text.ParseException;
import java.util.List;

/**
 * Factory that constructs {@link SenderClient}, {@link SubscriberClient} and {@link RouterClient}.
 */
public final class ClientFactory {

    private static final Logger LOGGER = LoggerFactory.getLogger(ClientFactory.class);

    private static final String SET_SCHEMA_DIR = "/set-schema";
    private static final String DESTINATION_SCHEMA_DIR = "/destination-schema";
    private static final String METADATA_SCHEMA_DIR = "/metadata-schema";
    private static final String BUILD_INFO_PATH = "buildinfo.yaml";
    private static final String CERTIFICATE_CONFIG_PATH = "certificate-config.yaml";

    private ClientFactory() {
    }

    /**
     * Create a new {@link SenderClient} to send submissions that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the sender client
     * @throws FitConnectInitialisationException on errors during the sender client construction
     */
    public static SenderClient getSenderClient(final ApplicationConfig config) throws FitConnectInitialisationException {

        checkNotNull(config);

        LOGGER.info("Initializing sender client ...");
        final SenderConfig senderConfig = config.getSenderConfig();
        final FitConnectService fitConnectService = createFitConnectService(config, senderConfig.getClientId(), senderConfig.getClientSecret());
        final ValidDataGuard dataGuard = new ValidDataGuard(fitConnectService);
        final SubmissionSender submissionSender = new SubmissionSender(config, fitConnectService, dataGuard);
        final ReplyReceiver replyReceiver = new ReplyReceiver(config, fitConnectService);

        return new SenderClient(fitConnectService, submissionSender, replyReceiver);
    }

    /**
     * Create a new {@link SubscriberClient} to receive submissions that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the subscriber client
     * @throws FitConnectInitialisationException on errors during the subscriber client construction
     */
    public static SubscriberClient getSubscriberClient(final ApplicationConfig config) throws FitConnectInitialisationException {

        checkNotNull(config);

        LOGGER.info("Initializing subscriber client ...");
        final SubscriberConfig subscriberConfig = config.getSubscriberConfig();

        final String decryptionKeyPath = getPrivateDecryptionKeyPathFromSubscriber(subscriberConfig);
        LOGGER.info("Reading private decryption key from {}", decryptionKeyPath);
        final String decryptionKey = readPath(decryptionKeyPath, "Decryption Key");
        final RSAKey rsaDecryptionKey = readRSAKeyFromString(decryptionKey);

        LOGGER.info("Reading private signing key from {} ", config.getSubscriberConfig().getPrivateSigningKeyPath());
        final String signingKey = readPath(config.getSubscriberConfig().getPrivateSigningKeyPath(), "Signing Key");
        final RSAKey rsaSigningKey = readRSAKeyFromString(signingKey);

        final FitConnectService fitConnectService = createFitConnectService(config, subscriberConfig.getClientId(), subscriberConfig.getClientSecret(), rsaSigningKey);
        final ValidDataGuard dataGuard = new ValidDataGuard(fitConnectService);

        final SubmissionReceiver submissionReceiver = new SubmissionReceiver(fitConnectService, rsaDecryptionKey, config);
        final ReplySender replySender = new ReplySender(config, fitConnectService, dataGuard);

        return new SubscriberClient(fitConnectService, submissionReceiver, replySender);
    }

    /**
     * Create a new {@link RouterClient} to find destinations and services that is automatically configured via a provided {@link ApplicationConfig}.
     *
     * @return the routing client
     * @throws FitConnectInitialisationException on errors during the routing client construction
     */
    public static RouterClient getRouterClient(final ApplicationConfig config) throws FitConnectInitialisationException {

        checkNotNull(config);

        LOGGER.info("Initializing router client ...");
        final HttpClient httpClient = getHttpClient(config);
        final SchemaProvider schemaProvider = getSchemaProvider(httpClient, config);
        final ValidationService validator = getValidationService(config, schemaProvider, new HashService());

        final KeyService keyService = new PublicKeyApiService(config, httpClient, validator);
        final RouteVerifier routeVerifier = new RouteVerifier(keyService, validator);
        final RoutingService routingService = new RoutingApiService(httpClient, config.getRoutingBaseUrl());

        return new RouterClient(routingService, routeVerifier);
    }


    private static FitConnectService createFitConnectService(final ApplicationConfig config, final String clientId, final String clientSecret) {
        return createFitConnectService(config, clientId, clientSecret, null);
    }

    private static FitConnectService createFitConnectService(final ApplicationConfig config, final String clientId, final String clientSecret, final RSAKey signingKey) {
        final HttpClient httpClient = getHttpClient(config);
        final SchemaProvider schemaProvider = getSchemaProvider(httpClient, config);
        final MessageDigestService messageDigestService = new HashService();

        final CryptoService cryptoService = new JWECryptoService(messageDigestService);
        final ValidationService validator = getValidationService(config, schemaProvider, messageDigestService);

        final OAuthService authService = new DefaultOAuthApiService(httpClient, clientId, clientSecret, config.getAuthBaseUrl());

        final SubmissionService submissionService = new SubmissionApiService(authService, httpClient, config.getSubmissionBaseUrl());
        final KeyService keyService = new PublicKeyApiService(config, httpClient, authService, submissionService, validator);
        final EventLogVerificationService eventLogVerifier = new EventLogVerifier(keyService, validator);
        final CaseService caseService = new CaseApiService(authService, httpClient, eventLogVerifier, config.getSubmissionBaseUrl());
        final ReplyService replyService = new ReplyApiService(authService, httpClient, config.getSubmissionBaseUrl());

        final SecurityEventService setService = new SecurityEventTokenService(config, validator, signingKey);

        return new FitConnectDefaultService(submissionService, caseService, replyService, cryptoService, validator, keyService, setService);
    }

    private static ValidationService getValidationService(final ApplicationConfig config, final SchemaProvider schemaProvider, final MessageDigestService messageDigestService) {
        final List<String> trustedRootCertificates = CertificateLoader.create().loadTrustedRootCertificates(loadCertificateConfig().getTrustedRootCertificates());
        LOGGER.info("Initialised trusted root certificates");
        return new DefaultValidationService(config, messageDigestService, schemaProvider, trustedRootCertificates);
    }

    private static HttpClient getHttpClient(final ApplicationConfig config) {
        final List<Interceptor> interceptors = List.of(new ApiRequestInterceptor(), new UserAgentInterceptor(loadBuildInfo()));
        return new HttpClient(config.getHttpConfig(), config.getProxyConfig(), interceptors);
    }

    private static SchemaProvider getSchemaProvider(final HttpClient httpClient, final ApplicationConfig config) {
        final List<String> setSchemaFiles = SchemaConfig.getSetSchemaFilePaths(SET_SCHEMA_DIR);
        final List<String> metadataSchemaFiles = SchemaConfig.getMetadataSchemaFileNames(METADATA_SCHEMA_DIR);
        final List<String> destinationSchemaFiles = SchemaConfig.getDestinationSchemaPaths(DESTINATION_SCHEMA_DIR);
        return new SchemaResourceProvider(httpClient, new SchemaResources(setSchemaFiles, metadataSchemaFiles, destinationSchemaFiles, config.getSubmissionDataSchemas(), config.getSubmissionDataSchemaPath()));
    }

    private static RSAKey readRSAKeyFromString(final String key) {
        try {
            return RSAKey.parse(JWK.parse(key).toJSONObject());
        } catch (final ParseException | NullPointerException e) {
            LOGGER.error("Could not parse key, please check environment (config.yml)");
            throw new FitConnectInitialisationException("Key could not be parsed", e);
        }
    }

    private static String readPath(final String pathName, final String errorKey) {
        try {
            return Files.readString(Path.of(pathName));
        } catch (final IOException | NullPointerException e) {
            LOGGER.error("Could not read {} from path '{}'. Please provide a valid path in your config.yml", errorKey, pathName);
            throw new FitConnectInitialisationException(e.getMessage(), e);
        }
    }

    private static BuildInfo loadBuildInfo() {
        try {
            final InputStream resource = ClientFactory.class.getClassLoader().getResourceAsStream(BUILD_INFO_PATH);
            return new Yaml().loadAs(resource, BuildInfo.class);
        } catch (final Exception e) {
            throw new FitConnectInitialisationException("Build info could not be loaded", e);
        }
    }

    private static CertificateConfig loadCertificateConfig() {
        try {
            final InputStream resource = ClientFactory.class.getClassLoader().getResourceAsStream(CERTIFICATE_CONFIG_PATH);
            return new Yaml().loadAs(resource, CertificateConfig.class);
        } catch (final Exception e) {
            throw new FitConnectInitialisationException("Certificate config could not be loaded", e);
        }
    }

    private static void checkNotNull(final ApplicationConfig config) {
        if (config == null) {
            throw new FitConnectInitialisationException("Application config must not be null");
        }
    }

    private static String getPrivateDecryptionKeyPathFromSubscriber(final SubscriberConfig subscriberConfig) {
        if (subscriberConfig.getPrivateDecryptionKeyPaths().size() != 1) {
            throw new FitConnectInitialisationException("Currently only one configured private key per subscriber is allowed !");
        }
        return subscriberConfig.getPrivateDecryptionKeyPaths().get(0);
    }
}
