package dev.fitko.fitconnect.client.sender;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.attachment.Attachment;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.domain.model.event.problems.attachment.AttachmentEncryptionIssue;
import dev.fitko.fitconnect.api.domain.model.event.problems.attachment.MissingAttachment;
import dev.fitko.fitconnect.api.domain.model.event.problems.data.DataEncryptionIssue;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.MetadataEncryptionIssue;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.MetadataJsonSyntaxViolation;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.InvalidEventLog;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.MissingAuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.submission.NotExactlyOneSubmitEvent;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.AttachmentForValidation;
import dev.fitko.fitconnect.api.domain.model.reply.Reply;
import dev.fitko.fitconnect.api.domain.sender.ReceivedReply;
import dev.fitko.fitconnect.api.domain.subscriber.ReceivedReplyData;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectReplyException;
import dev.fitko.fitconnect.api.exceptions.internal.AuthenticationTagsEmptyException;
import dev.fitko.fitconnect.api.exceptions.internal.DecryptionException;
import dev.fitko.fitconnect.api.exceptions.internal.EventLogException;
import dev.fitko.fitconnect.api.exceptions.internal.RestApiException;
import dev.fitko.fitconnect.api.exceptions.internal.SubmissionRequestException;
import dev.fitko.fitconnect.api.exceptions.internal.SubmitEventNotFoundException;
import dev.fitko.fitconnect.core.util.StopWatch;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.stream.Collectors;

import static com.fasterxml.jackson.databind.DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES;

public class ReplyReceiver {

    private static final Logger LOGGER = LoggerFactory.getLogger(ReplyReceiver.class);

    private static final ObjectMapper MAPPER = new ObjectMapper().configure(FAIL_ON_UNKNOWN_PROPERTIES, false);

    private final ApplicationConfig config;
    private final FitConnectService fitConnectService;

    public ReplyReceiver(final ApplicationConfig config, final FitConnectService fitConnectService) {
        this.config = config;
        this.fitConnectService = fitConnectService;
    }

    public ReceivedReply receiveReply(final UUID replyId, final RSAKey replyDecryptionKey) {

        LOGGER.info("Requesting reply ...");
        final Reply reply = loadReply(replyId);

        LOGGER.info("Loading authentication tags from event log ...");
        final AuthenticationTags authenticationTagsFromEvent = loadAuthTagsForSubmitEvent(reply);

        LOGGER.info("Decrypting reply metadata...");
        final Metadata metadata = decryptMetadata(replyDecryptionKey, reply);
        validateMetadata(reply, authenticationTagsFromEvent, metadata);

        LOGGER.info("Decrypting reply data...");
        final byte[] decryptedData = decryptData(replyDecryptionKey, reply);
        validateData(reply, authenticationTagsFromEvent, metadata, decryptedData);

        LOGGER.info("Loading reply attachments");
        final List<AttachmentForValidation> attachments = loadAttachments(replyDecryptionKey, reply, metadata);
        final AuthenticationTags authenticationTags = buildAuthenticationTags(reply, attachments);
        validateAttachments(reply, attachments, authenticationTags);

        return buildReceivedReply(reply, metadata, decryptedData, attachments, authenticationTags);
    }

    private void validateAttachments(final Reply reply, final List<AttachmentForValidation> attachments, final AuthenticationTags authenticationTags) {
        final ValidationResult validatedAttachments = fitConnectService.validateAttachments(attachments, authenticationTags);
        if (!validatedAttachments.isValid()) {
            LOGGER.error("Attachment validation failed", validatedAttachments.getError());
            evaluateValidationResult(reply, validatedAttachments);
        }
    }

    private void validateData(final Reply reply, final AuthenticationTags authenticationTagsFromEvent, final Metadata metadata, final byte[] decryptedData) {
        final ValidationResult validatedData = fitConnectService.validateData(decryptedData, reply.getEncryptedData(), metadata, authenticationTagsFromEvent.getData());
        if (!validatedData.isValid()) {
            LOGGER.error("Data validation failed", validatedData.getError());
            evaluateValidationResult(reply, validatedData);
        }
    }

    private void validateMetadata(final Reply reply, final AuthenticationTags authenticationTagsFromEvent, final Metadata metadata) {
        final ValidationResult validatedMetadata = fitConnectService.validateReplyMetadata(metadata, reply, authenticationTagsFromEvent);
        if (!validatedMetadata.isValid()) {
            LOGGER.error("Metadata validation failed", validatedMetadata.getError());
            evaluateValidationResult(reply, validatedMetadata);
        }
    }

    private Reply loadReply(final UUID replyId) {
        try {
            return fitConnectService.getReply(replyId);
        } catch (final RestApiException e) {
            throw new FitConnectReplyException(e.getMessage(), e);
        }
    }

    private AuthenticationTags loadAuthTagsForSubmitEvent(final Reply reply) {
        try {
            return fitConnectService.getReplyAuthenticationTags(reply);
        } catch (final EventLogException e) {
            // https://docs.fitko.de/fit-connect/docs/receiving/verification/#struktur--und-signaturpr%C3%BCfung-der-security-event-tokens
            final InvalidEventLog problem = new InvalidEventLog();
            reject(reply, List.of(problem));
            throw new FitConnectReplyException(problem.getDetail(), e);
        } catch (final SubmitEventNotFoundException e) {
            // https://docs.fitko.de/fit-connect/docs/receiving/verification/#genau-ein-submit-event
            final NotExactlyOneSubmitEvent problem = new NotExactlyOneSubmitEvent();
            reject(reply, List.of(problem));
            throw new FitConnectReplyException(problem.getDetail());
        } catch (final AuthenticationTagsEmptyException e) {
            // https://docs.fitko.de/fit-connect/docs/receiving/verification/#authentication-tags-im-submit-event
            final MissingAuthenticationTags problem = new MissingAuthenticationTags();
            reject(reply, List.of(problem));
            throw new FitConnectReplyException(problem.getDetail());
        }
    }

    private static AuthenticationTags buildAuthenticationTags(final Reply reply, final List<AttachmentForValidation> attachments) {
        final Map<UUID, String> encryptedAttachments = attachments.stream().collect(Collectors.toMap(AttachmentForValidation::getAttachmentId, AttachmentForValidation::getEncryptedData));
        return AuthenticationTags.from(reply.getEncryptedData(), reply.getEncryptedMetadata(), encryptedAttachments);
    }

    private ReceivedReply buildReceivedReply(final Reply reply, final Metadata metadata, final byte[] decryptedData, final List<AttachmentForValidation> attachments, final AuthenticationTags authenticationTags) {
        final String data = new String(decryptedData, StandardCharsets.UTF_8);
        final List<Attachment> receivedAttachments = attachments.stream().map(this::toApiAttachment).collect(Collectors.toList());
        final ReceivedReplyData receivedData = new ReceivedReplyData(data, metadata, reply, receivedAttachments, authenticationTags );
        return new ReceivedReply(fitConnectService, receivedData);
    }

    private Attachment toApiAttachment(final AttachmentForValidation attachment) {
        final ApiAttachment metadata = attachment.getAttachmentMetadata();
        return Attachment.fromByteArray(attachment.getDecryptedData(), metadata.getMimeType(), metadata.getFilename(), metadata.getDescription());
    }

    private List<AttachmentForValidation> loadAttachments(final RSAKey decryptionKey, final Reply reply, final Metadata metadata) {
        final List<ApiAttachment> attachments = metadata.getContentStructure().getAttachments();
        if (attachments == null || attachments.isEmpty()) {
            LOGGER.info("Reply contains no attachments");
            return Collections.emptyList();
        }
        final List<AttachmentForValidation> receivedAttachments = new ArrayList<>();
        for (final ApiAttachment attachmentMetadata : attachments) {
            final String encryptedAttachment = downloadAttachment(reply, attachmentMetadata);
            final byte[] decryptedAttachment = decryptAttachment(reply, decryptionKey, attachmentMetadata, encryptedAttachment);
            receivedAttachments.add(new AttachmentForValidation(attachmentMetadata, encryptedAttachment, decryptedAttachment));
        }
        return receivedAttachments;
    }

    private String downloadAttachment(final Reply reply, final ApiAttachment metadata) {
        try {
            return fitConnectService.getReplyAttachment(reply.getReplyId(), metadata.getAttachmentId());
        } catch (final RestApiException e) {
            if (e.isNotFound()) {
                reject(reply, List.of(new MissingAttachment(metadata.getAttachmentId())));
            }
            throw new FitConnectReplyException(e.getMessage(), e);
        }
    }

    private byte[] decryptAttachment(final Reply reply, final RSAKey decryptionKey, final ApiAttachment metadata, final String encryptedAttachment) {
        try {
            final var startDecryption = StopWatch.start();
            final byte[] decryptedAttachment = fitConnectService.decryptString(decryptionKey, encryptedAttachment);
            LOGGER.trace("Decrypting attachment {} took {}", metadata.getAttachmentId(), StopWatch.stop(startDecryption));
            return decryptedAttachment;
        } catch (final DecryptionException e) {
            // https://docs.fitko.de/fit-connect/docs/receiving/verification/#entschl%C3%BCsselung-2
            reject(reply, List.of(new AttachmentEncryptionIssue(metadata.getAttachmentId())));
            throw new FitConnectReplyException(e.getMessage(), e);
        }
    }

    private byte[] decryptData(final RSAKey replyDecryptionKey, final Reply reply) {
        try {
            return fitConnectService.decryptString(replyDecryptionKey, reply.getEncryptedData());
        } catch (final DecryptionException e) {
            // https://docs.fitko.de/fit-connect/docs/receiving/verification/#entschl%C3%BCsselung-1
            reject(reply, List.of(new DataEncryptionIssue()));
            throw new FitConnectReplyException(e.getMessage(), e);
        }
    }

    private Metadata decryptMetadata(final RSAKey replyDecryptionKey, final Reply reply) {
        try {
            final byte[] metadataBytes = fitConnectService.decryptString(replyDecryptionKey, reply.getEncryptedMetadata());
            return MAPPER.readValue(metadataBytes, Metadata.class);
        } catch (final IOException e) {
            reject(reply, List.of(new MetadataJsonSyntaxViolation()));
            throw new FitConnectReplyException(e.getMessage(), e);
        } catch (final DecryptionException e) {
            // https://docs.fitko.de/fit-connect/docs/receiving/verification/#entschl%C3%BCsselung
            reject(reply, List.of(new MetadataEncryptionIssue()));
            throw new FitConnectReplyException(e.getMessage(), e);
        }
    }

    private void evaluateValidationResult(final Reply reply, final ValidationResult validationResult) throws SubmissionRequestException {
        if (validationResult.hasProblems()) {
            reject(reply, validationResult.getProblems());
            throw new FitConnectReplyException(validationResult.getProblems().stream().map(Problem::getDetail).collect(Collectors.joining()), validationResult.getError());
        }
        LOGGER.error(validationResult.getError().getMessage(), validationResult.getError());
        throw new FitConnectReplyException(validationResult.getError().getMessage(), validationResult.getError());
    }

    private void reject(final Reply reply, final List<Problem> problems) {
        if (config.isAutoRejectEnabled()) {
            LOGGER.info("Auto-rejecting reply due the following problem(s): {}", problems.stream().map(Problem::getDetail).collect(Collectors.joining("\n")));
            fitConnectService.rejectReply(reply.getReplyId(), problems);
        }
    }

}
