package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.domain.model.attachment.AttachmentPayload;
import dev.fitko.fitconnect.api.domain.model.callback.Callback;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.ApiAttachment;
import dev.fitko.fitconnect.api.domain.model.metadata.attachment.Purpose;
import dev.fitko.fitconnect.api.domain.model.submission.AnnounceSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.SentSubmission;
import dev.fitko.fitconnect.api.domain.model.submission.ServiceType;
import dev.fitko.fitconnect.api.domain.model.submission.Submission;
import dev.fitko.fitconnect.api.domain.model.submission.SubmitSubmission;
import dev.fitko.fitconnect.api.domain.sender.SendableEncryptedSubmission;
import dev.fitko.fitconnect.api.domain.sender.SendableSubmission;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.stream.Collectors;

/**
 * Helper to create types for announced, sent and submit submissions.
 */
public final class SubmissionUtil {

    private SubmissionUtil() {

    }

    public static AnnounceSubmission buildCreateSubmission(final SendableSubmission sendableSubmission, final List<AttachmentPayload> attachments) {
        final List<UUID> attachmentIds = attachments.stream().map(AttachmentPayload::getAttachmentId).collect(Collectors.toList());
        final ServiceType serviceType = getServiceTypeFromSendableSubmission(sendableSubmission.getServiceName(), sendableSubmission.getServiceIdentifier());
        return getCreateSubmission(sendableSubmission.getDestinationId(), sendableSubmission.getCaseId(), attachmentIds, serviceType, sendableSubmission.getCallback());
    }

    public static AnnounceSubmission buildCreateSubmission(final SendableEncryptedSubmission sendableSubmission, final Map<UUID, String> attachments) {
        final ServiceType serviceType = getServiceTypeFromSendableSubmission(sendableSubmission.getServiceName(), sendableSubmission.getServiceIdentifier());
        return getCreateSubmission(sendableSubmission.getDestinationId(), sendableSubmission.getCaseId(), new ArrayList<>(attachments.keySet()), serviceType, sendableSubmission.getCallback());
    }

    private static AnnounceSubmission getCreateSubmission(final UUID destinationId, final UUID caseId, final List<UUID> attachments, final ServiceType serviceType, final Callback callback) {
        return AnnounceSubmission.builder()
                .destinationId(destinationId)
                .caseId(caseId)
                .announcedAttachments(attachments)
                .callback(callback)
                .serviceType(serviceType)
                .build();
    }

    private static ServiceType getServiceTypeFromSendableSubmission(final String serviceName, final String serviceIdentifier) {
        return ServiceType.builder()
                .name(serviceName)
                .identifier(serviceIdentifier)
                .build();
    }

    public static SubmitSubmission buildSubmitSubmission(final UUID submissionId, final String encryptedData, final String encryptedMetadata) {
        final SubmitSubmission submission = new SubmitSubmission();
        submission.setSubmissionId(submissionId);
        submission.setEncryptedData(encryptedData);
        submission.setEncryptedMetadata(encryptedMetadata);
        return submission;
    }

    public static SentSubmission buildSentSubmission(final AuthenticationTags authenticationTags, final Submission submission) {
        return SentSubmission.builder()
                .submissionId(submission.getSubmissionId())
                .caseId(submission.getCaseId())
                .destinationId(submission.getDestinationId())
                .authenticationTags(authenticationTags)
                .build();
    }

    public static ApiAttachment toApiAttachment(final AttachmentPayload attachmentPayload) {
        final var attachment = new ApiAttachment();
        attachment.setAttachmentId(attachmentPayload.getAttachmentId());
        attachment.setPurpose(Purpose.ATTACHMENT);
        attachment.setFilename(attachmentPayload.getFileName());
        attachment.setMimeType(attachmentPayload.getMimeType());
        attachment.setDescription(attachmentPayload.getDescription());

        final var hash = new Hash();
        hash.setContent(attachmentPayload.getHashedData());
        hash.setSignatureType(SignatureType.SHA_512);
        attachment.setHash(hash);
        return attachment;
    }
}
