package dev.fitko.fitconnect.client.util;

import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.domain.model.callback.Callback;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.destination.DestinationService;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.FitConnect;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.domain.sender.SendableEncryptedSubmission;
import dev.fitko.fitconnect.api.domain.sender.SendableReply;
import dev.fitko.fitconnect.api.domain.sender.SendableSubmission;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;

import java.net.URI;
import java.util.HashSet;
import java.util.Objects;
import java.util.Optional;
import java.util.Set;
import java.util.UUID;
import java.util.function.Predicate;
import java.util.regex.Pattern;

public class ValidDataGuard {

    private static final Pattern LEIKA_KEY_PATTERN = Pattern.compile("^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$");
    private final FitConnectService sender;

    public ValidDataGuard(final FitConnectService sender) {
        this.sender = sender;
    }

    /**
     * Checks if the encrypted submission payload is in a valid state for sending.
     *
     * @param sendableEncryptedSubmission payload to be checked
     * @throws IllegalArgumentException if one of the checks fails
     */
    public void ensureValidDataPayload(final SendableEncryptedSubmission sendableEncryptedSubmission) {
        if (sendableEncryptedSubmission == null) {
            throw new IllegalArgumentException("Encrypted payload must not be null.");
        }
        if (sendableEncryptedSubmission.getData() == null) {
            throw new IllegalArgumentException("Encrypted data is mandatory, but was null.");
        }
        if (sendableEncryptedSubmission.getMetadata() == null) {
            throw new IllegalArgumentException("Encrypted metadata must not be null.");
        }
        testDefaults(sendableEncryptedSubmission.getDestinationId(), sendableEncryptedSubmission.getServiceIdentifier(), sendableEncryptedSubmission.getCallback());
    }

    /**
     * Checks if the unencrypted submission payload is in a valid state for sending.
     *
     * @param sendableSubmission payload to be checked
     * @throws IllegalArgumentException if one of the checks fails
     */
    public void ensureValidDataPayload(final SendableSubmission sendableSubmission) {
        if (sendableSubmission == null) {
            throw new IllegalArgumentException("Payload must not be null.");
        }
        if (sendableSubmission.getData() == null) {
            throw new IllegalArgumentException("Data is mandatory, but was null.");
        }
        testOnValidDataFormat(sendableSubmission.getData(), sendableSubmission.getSubmissionSchema());
        testDefaults(sendableSubmission);
    }

    /**
     * Checks if the unencrypted reply payload is in a valid state for sending.
     *
     * @param sendableReply reply payload to be checked
     * @throws IllegalArgumentException if one of the checks fails
     */
    public void ensureValidDataPayload(final SendableReply sendableReply) {
        if (sendableReply == null) {
            throw new IllegalArgumentException("Payload must not be null.");
        } else if (sendableReply.getData() == null) {
            throw new IllegalArgumentException("Data is mandatory, but was null.");
        } else if (sendableReply.getReplyChannel() == null || sendableReply.getReplyChannel().getFitConnect() == null) {
            throw new IllegalArgumentException("FitConnect reply channel is mandatory for sending replies, but was null on original submission " + sendableReply.getOriginalSubmissionId());
        }

        final Destination destination = sender.getDestination(sendableReply.getDestinationId());
        checkMatchingDataSchemaOnDestination(sendableReply.getSubmissionSchema(), destination);
        testOnValidDataFormat(sendableReply.getData(), sendableReply.getSubmissionSchema());
        checkMatchingProcessingStandardsOnDestination(destination.getServices(), sendableReply.getReplyChannel(), sendableReply.getServiceIdentifier(), sendableReply.getSubmissionSchema());
    }

    private void testDefaults(final UUID destinationId, final String serviceIdentifier, final Callback callback) {
        if (destinationId == null) {
            throw new IllegalArgumentException("DestinationId is mandatory, but was null.");
        } else if (serviceIdentifier == null) {
            throw new IllegalArgumentException("Leika key is mandatory, but was null.");
        } else if (noValidLeikaKeyPattern(serviceIdentifier)) {
            throw new IllegalArgumentException("LeikaKey has invalid format, please follow: ^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$.");
        }

        final Destination destination = sender.getDestination(destinationId);
        if (serviceTypeDoesNotMatchDestination(destination, serviceIdentifier)) {
            throw new IllegalArgumentException("Provided service type '" + serviceIdentifier + "' is not allowed by the destination ");
        }
        if (callback != null) {
            checkCallbackFormat(callback);
        }
    }

    private void testDefaults(final SendableSubmission sendableSubmission) {
        if (sendableSubmission.getDestinationId() == null) {
            throw new IllegalArgumentException("DestinationId is mandatory, but was null.");
        } else if (sendableSubmission.getServiceIdentifier() == null) {
            throw new IllegalArgumentException("Leika key is mandatory, but was null.");
        } else if (noValidLeikaKeyPattern(sendableSubmission.getServiceIdentifier())) {
            throw new IllegalArgumentException("LeikaKey has invalid format, please follow: ^urn:[a-z0-9][a-z0-9-]{0,31}:[a-z0-9()+,.:=@;$_!*'%/?#-]+$.");
        }

        final Destination destination = sender.getDestination(sendableSubmission.getDestinationId());
        if (serviceTypeDoesNotMatchDestination(destination, sendableSubmission.getServiceIdentifier())) {
            throw new IllegalArgumentException("Provided service type '" + sendableSubmission.getServiceIdentifier() + "' is not allowed by the destination ");
        }
        checkMatchingDataSchemaOnDestination(sendableSubmission.getSubmissionSchema(), destination);
        checkMatchingProcessingStandardsOnDestination(destination.getServices(), sendableSubmission.getReplyChannel(), sendableSubmission.getServiceIdentifier(), sendableSubmission.getSubmissionSchema());
        if (sendableSubmission.getCallback() != null) {
            checkCallbackFormat(sendableSubmission.getCallback());
        }
    }

    private void checkMatchingProcessingStandardsOnDestination(final Set<DestinationService> destinationServices, ReplyChannel submissionReplyChannel, String serviceIdentifier, final SubmissionSchema submissionSchema) {
        if (processingStandardsDoNotMatchDestination(destinationServices, submissionReplyChannel, serviceIdentifier, submissionSchema)) {
            throw new IllegalArgumentException("FIT-Connect reply channel processing standard(s) do not match any of the destination services");
        }
    }

    private boolean processingStandardsDoNotMatchDestination(final Set<DestinationService> destinationServices, ReplyChannel submissionReplyChannel, String serviceIdentifier, final SubmissionSchema submissionSchema) {
        if (destinationServices == null
                || destinationServices.isEmpty()
                || destinationServices.stream().noneMatch(s -> s.getReplyChannels() != null)
                || submissionReplyChannel == null
                || submissionReplyChannel.getFitConnect() == null) {
            return false;
        }

        final Optional<FitConnect> matchingStandardsFound = destinationServices.stream()
                .filter(service -> service.getIdentifier().equals(serviceIdentifier))
                .filter(service -> service.getSubmissionSchemas().stream().anyMatch(schema -> schema.equals(submissionSchema)))
                .map(DestinationService::getReplyChannels)
                .map(ReplyChannel::getFitConnect)
                .filter(Objects::nonNull)
                .filter(matchingProcessingStandards(submissionReplyChannel))
                .findAny();

        return matchingStandardsFound.isEmpty();
    }

    private static Predicate<FitConnect> matchingProcessingStandards(ReplyChannel submissionReplyChannel) {
        return fitConnect -> new HashSet<>(fitConnect.getProcessStandards()).containsAll(submissionReplyChannel.getFitConnect().getProcessStandards());
    }

    private void checkMatchingDataSchemaOnDestination(final SubmissionSchema submissionSchema, final Destination destination) {
        final MimeType dataMimeType = submissionSchema.getMimeType();
        if (mimeTypeAndSchemaUriDoNotMatchDestination(destination, dataMimeType, submissionSchema.getSchemaUri())) {
            throw new IllegalArgumentException("Combination of provided MIME type '" + dataMimeType.value() + "' and schema URI '" + submissionSchema.getSchemaUri() + "' is not allowed by the destination");
        }
    }

    private boolean serviceTypeDoesNotMatchDestination(final Destination destination, final String serviceIdentifier) {
        return destination.getServices().stream()
                .map(DestinationService::getIdentifier)
                .filter(Objects::nonNull)
                .filter(destinationServiceIdentifier -> destinationServiceIdentifier.equals(serviceIdentifier))
                .findFirst()
                .isEmpty();
    }

    private boolean mimeTypeAndSchemaUriDoNotMatchDestination(final Destination destination, final MimeType mimeType, final URI schemaUri) {
        return destination.getServices().stream()
                .flatMap(service -> service.getSubmissionSchemas().stream())
                .filter(Objects::nonNull)
                .filter(schema -> schema.getMimeType().equals(mimeType))
                .filter(schema -> schema.getSchemaUri().equals(schemaUri))
                .findFirst()
                .isEmpty();
    }

    private void testOnValidDataFormat(final String data, SubmissionSchema submissionSchema) {
        final MimeType dataMimeType = submissionSchema.getMimeType();
        if (dataMimeType.equals(MimeType.APPLICATION_JSON)) {
            checkJsonFormat(data, submissionSchema.getSchemaUri());
        } else if (dataMimeType.equals(MimeType.APPLICATION_XML)) {
            checkXmlFormat(data);
        }
    }

    private void checkXmlFormat(final String xmlData) {
        final ValidationResult validationResult = sender.validateXmlFormat(xmlData);
        if (validationResult.hasError()) {
            throw new IllegalArgumentException("Data is not in expected xml format, please provide valid xml: " + validationResult.getError().getMessage());
        }
    }

    private void checkJsonFormat(final String jsonData, final URI dataSchemaUri) {
        final ValidationResult validationResult = sender.validateJsonFormat(jsonData, dataSchemaUri);
        if (validationResult.hasError()) {
            throw new IllegalArgumentException("Data is not in expected json format, please provide valid json: " + validationResult.getError().getMessage());
        }
    }

    private void checkCallbackFormat(final Callback callback) {
        if (callback.getUri() == null || !callback.getUri().getScheme().equals("https")) {
            throw new IllegalArgumentException("Callback URI " + callback.getUri() + " must be a secure https connection");
        }
        if (callback.getSecret() == null || callback.getSecret().length() < 32) {
            throw new IllegalArgumentException("Callback secret must have a length of at least 32 characters");
        }
    }

    private boolean noValidLeikaKeyPattern(final String leikaKey) {
        if (leikaKey.length() < 7 || leikaKey.length() > 255) {
            return true;
        }
        return !LEIKA_KEY_PATTERN.matcher(leikaKey).matches();
    }
}
