package dev.fitko.fitconnect.client.sender;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.DeserializationFeature;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.FitConnectService;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.domain.crypto.JWKPair;
import dev.fitko.fitconnect.api.domain.model.event.authtags.AuthenticationTags;
import dev.fitko.fitconnect.api.domain.model.event.problems.metadata.MetadataSchemaViolation;
import dev.fitko.fitconnect.api.domain.model.metadata.ContentStructure;
import dev.fitko.fitconnect.api.domain.model.metadata.Hash;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.metadata.SignatureType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.Data;
import dev.fitko.fitconnect.api.domain.model.metadata.data.MimeType;
import dev.fitko.fitconnect.api.domain.model.metadata.data.SubmissionSchema;
import dev.fitko.fitconnect.api.domain.model.reply.Reply;
import dev.fitko.fitconnect.api.domain.model.reply.replychannel.ReplyChannel;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.client.FitConnectReplyException;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.client.util.ReplyChannelKeyGenerator;
import dev.fitko.fitconnect.core.crypto.HashService;
import dev.fitko.fitconnect.core.crypto.JWECryptoService;
import dev.fitko.fitconnect.core.util.EventLogUtil;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.net.URI;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertNotNull;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.times;
import static org.mockito.Mockito.verify;
import static org.mockito.Mockito.when;

class ReplyReceiverTest {

    private static final ObjectMapper MAPPER = new ObjectMapper().configure(DeserializationFeature.FAIL_ON_UNKNOWN_PROPERTIES, false);

    private FitConnectService serviceMock;

    private ReplyReceiver underTest;

    @BeforeEach
    void setUp() {

        final EnvironmentName envName = new EnvironmentName("TEST_ENV");
        final Environment testEnv = new Environment();
        testEnv.setEnableAutoReject(true);

        final ApplicationConfig config = ApplicationConfig.builder()
                .activeEnvironment(envName)
                .environments(Map.of(envName, testEnv))
                .build();

        serviceMock = mock(FitConnectService.class);
        underTest = new ReplyReceiver(config, serviceMock);
    }

    @Test
    void testReceiveValidReply() throws JsonProcessingException {

        // Given
        final var replyId = UUID.randomUUID();

        final JWKPair keyPair = ReplyChannelKeyGenerator.generateKeyPair();
        final RSAKey replyEncryptionKey = keyPair.getPublicKey().toRSAKey();
        final RSAKey replyDecryptionKey = keyPair.getPrivateKey().toRSAKey();

        final var dataPayload = "{ some : 'data foo' }";
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Hash hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        hash.setSignatureType(SignatureType.SHA_512);

        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://dummy.schema.url"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(List.of());
        contentStructure.setData(data);

        final Metadata metadata = new Metadata();
        metadata.setContentStructure(contentStructure);
        metadata.setReplyChannel(ReplyChannel.fromFitConnect(replyEncryptionKey, List.of("https://processing.standard.net")));

        final byte[] metadataBytes = MAPPER.writeValueAsBytes(metadata);

        final String encryptedData = cryptoService.encryptObject(replyEncryptionKey, dataPayload, MimeType.APPLICATION_JSON.value());
        final String encryptedMetadata = cryptoService.encryptBytes(replyEncryptionKey, metadataBytes, MimeType.APPLICATION_JSON.value());

        final Reply expectedReply = new Reply();
        expectedReply.setCaseId(UUID.randomUUID());
        expectedReply.setReplyId(replyId);
        expectedReply.setEncryptedData(encryptedData);
        expectedReply.setEncryptedMetadata(encryptedMetadata);

        final AuthenticationTags replySubmitEventAuthTags = new AuthenticationTags();
        replySubmitEventAuthTags.setData(encryptedData.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);
        replySubmitEventAuthTags.setMetadata(encryptedMetadata.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);

        when(serviceMock.validateReplyMetadata(any(), any(), any())).thenReturn(ValidationResult.ok());
        when(serviceMock.getReply(replyId)).thenReturn(expectedReply);
        when(serviceMock.getReplyAuthenticationTags(expectedReply)).thenReturn(replySubmitEventAuthTags);
        when(serviceMock.validateData(any(), any(), any(), any())).thenReturn(ValidationResult.ok());
        when(serviceMock.validateAttachments(any(), any())).thenReturn(ValidationResult.ok());
        when(serviceMock.validateJsonFormat(any(), any())).thenReturn(ValidationResult.ok());
        when(serviceMock.decryptString(replyDecryptionKey, encryptedData)).thenReturn(dataPayload.getBytes());
        when(serviceMock.decryptString(replyDecryptionKey, encryptedMetadata)).thenReturn(metadataBytes);

        // When
        final var receivedReply = underTest.receiveReply(replyId, replyDecryptionKey);

        // Then
        assertNotNull(receivedReply);
        assertThat(receivedReply.getReplyId(), is(replyId));
        assertThat(receivedReply.getDataAsString(), is(dataPayload));
        assertThat(receivedReply.getDataMimeType(), is(MimeType.APPLICATION_JSON.value()));
        assertThat(receivedReply.getAttachments(), is(Collections.EMPTY_LIST));

    }

    @Test
    void testAutoRejectInvalidReply() throws JsonProcessingException {

        // Given
        final var replyId = UUID.randomUUID();

        final JWKPair keyPair = ReplyChannelKeyGenerator.generateKeyPair();
        final RSAKey replyEncryptionKey = keyPair.getPublicKey().toRSAKey();
        final RSAKey replyDecryptionKey = keyPair.getPrivateKey().toRSAKey();

        final var dataPayload = "{ some : 'data foo' }";
        final CryptoService cryptoService = new JWECryptoService(new HashService());

        final Hash hash = new Hash();
        hash.setContent(cryptoService.hashBytes(dataPayload.getBytes()));
        hash.setSignatureType(SignatureType.SHA_512);

        final SubmissionSchema schema = new SubmissionSchema();
        schema.setSchemaUri(URI.create("https://dummy.schema.url"));
        schema.setMimeType(MimeType.APPLICATION_JSON);

        final Data data = new Data();
        data.setSubmissionSchema(schema);
        data.setHash(hash);

        final ContentStructure contentStructure = new ContentStructure();
        contentStructure.setAttachments(List.of());
        contentStructure.setData(data);

        final Metadata metadata = new Metadata();
        metadata.setContentStructure(contentStructure);
        metadata.setReplyChannel(ReplyChannel.fromFitConnect(replyEncryptionKey, List.of("https://processing.standard.net")));

        final byte[] metadataBytes = MAPPER.writeValueAsBytes(metadata);

        final String encryptedData = cryptoService.encryptObject(replyEncryptionKey, dataPayload, MimeType.APPLICATION_JSON.value());
        final String encryptedMetadata = cryptoService.encryptBytes(replyEncryptionKey, metadataBytes, MimeType.APPLICATION_JSON.value());

        final Reply expectedReply = new Reply();
        expectedReply.setCaseId(UUID.randomUUID());
        expectedReply.setReplyId(replyId);
        expectedReply.setEncryptedData(encryptedData);
        expectedReply.setEncryptedMetadata(encryptedMetadata);

        final AuthenticationTags replySubmitEventAuthTags = new AuthenticationTags();
        replySubmitEventAuthTags.setData(encryptedData.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);
        replySubmitEventAuthTags.setMetadata(encryptedMetadata.split(EventLogUtil.AUTH_TAG_SPLIT_TOKEN)[4]);

        when(serviceMock.validateReplyMetadata(any(), any(), any())).thenReturn(ValidationResult.problem(new MetadataSchemaViolation()));
        when(serviceMock.getReply(replyId)).thenReturn(expectedReply);
        when(serviceMock.getReplyAuthenticationTags(expectedReply)).thenReturn(replySubmitEventAuthTags);
        when(serviceMock.decryptString(replyDecryptionKey, encryptedMetadata)).thenReturn(metadataBytes);

        // When
        final FitConnectReplyException exception = assertThrows(FitConnectReplyException.class, () -> underTest.receiveReply(replyId, replyDecryptionKey));

        // Then
        assertThat(exception.getMessage(), containsString("Submission metadata does not comply to schema"));
        verify(serviceMock, times(1)).rejectReply(replyId, List.of(new MetadataSchemaViolation()));

    }
}